from .core import spell, no_spells

from .spells import unpack_keys, unpack_attrs, args_with_source, dict_of, print_args, call_with_name, delegate_to_attr, \
    maybe, select_from, magic_kwargs, assigned_names, switch, timeit

__version__ = '0.2.1'
